/*
 * File Name:test.c - Phytium driver test
 *
 * Copyright (C) 2020 Phytium Technology Co.,Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#include <linux/module.h>
#include <linux/irq.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/cdev.h>
#include <linux/types.h>
#include <linux/mutex.h>
#include <linux/acpi.h>
#include <crypto/internal/hash.h>
#include <linux/crypto.h>
#include <crypto/hash.h>
#include <crypto/internal/skcipher.h>
#include <crypto/internal/akcipher.h>
#include <linux/ioport.h>
#include <linux/io.h>
#include "sm4_software.h"
#include "sm3_software.h"


char *alg = NULL;
u32 slen = 256;
u32 elen = 2048;

module_param(alg, charp, 0);
module_param(slen, uint, 0);
module_param(elen, uint, 0);

#define get_driver_name(tfm_type, tfm) crypto_tfm_alg_driver_name(tfm_type ## _tfm(tfm))


static void sm3_dgst_test(const u8 *data, unsigned int len, u8 *out)
{
	struct scatterlist sg;
	struct crypto_wait wait;
	struct ahash_request *req;
	struct crypto_ahash *tfm;
	int ret;
	static u32 flag = 0;

	tfm = crypto_alloc_ahash("sm3", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", "sm3", PTR_ERR(tfm));
		return;
	}

	if (flag++ == 0)
		pr_info( "\ntesting of async %s (%s)\n", "sm3", get_driver_name(crypto_ahash, tfm));

	sg_init_table(&sg, 1);
	sg_set_buf(&sg, data, len);

	req = ahash_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("ahash request allocation failure\n");
		goto fail;
	}

	crypto_init_wait(&wait);
	ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);

	ahash_request_set_crypt(req, &sg, out, len);

	ret = crypto_wait_req(crypto_ahash_digest(req), &wait);
	if (ret) {
		pr_err("hashing failed ret=%d\n", ret);
	}

	ahash_request_free(req);
fail:
	crypto_free_ahash(tfm);
}

static void sm3_step_test(const u8 *data, unsigned int len, u8 *out)
{
	struct scatterlist sg;
	struct crypto_wait wait;
	struct ahash_request *req;
	struct crypto_ahash *tfm;
	int ret;
	unsigned int offset, step_len;
	static u32 flag = 0;

	tfm = crypto_alloc_ahash("sm3", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", "sm3", PTR_ERR(tfm));
		return;
	}

	if (flag++ == 0)
		pr_info( "\ntesting of async %s (%s)\n", "sm3", get_driver_name(crypto_ahash, tfm));

	req = ahash_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("ahash request allocation failure\n");
		goto fail;
	}

	crypto_init_wait(&wait);
	ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);
	ret = crypto_wait_req(crypto_ahash_init(req), &wait);
	if (ret) {
		pr_err("hashing failed ret=%d\n", ret);
		goto fail1;
	}

	for (offset = 0; offset < len;) {
		step_len = 0xFF;
		if (offset + step_len > len)
			step_len = len - offset;

		crypto_init_wait(&wait);
		ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);

		sg_init_table(&sg, 1);
		sg_set_buf(&sg, data + offset, step_len);
		ahash_request_set_crypt(req, &sg, out, step_len);

		ret = crypto_wait_req(crypto_ahash_update(req), &wait);
		if (ret) {
			pr_err("crypto_ahash_update failed ret=%d\n", ret);
			break;
		}
		offset += step_len;
	}

	crypto_init_wait(&wait);
	ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);

	sg_init_table(&sg, 1);
	sg_set_buf(&sg, data, 0);
	ahash_request_set_crypt(req, &sg, out, 0);
	ret = crypto_wait_req(crypto_ahash_final(req), &wait);
	if (ret) {
		pr_err("crypto_ahash_final failed ret=%d\n", ret);
	}

fail1:
	ahash_request_free(req);
fail:
	crypto_free_ahash(tfm);
}

static void hmac_sm3_test(const u8 *key, unsigned int klen, const u8 *data, unsigned int len, u8 *out)
{
	struct scatterlist sg;
	struct crypto_wait wait;
	struct ahash_request *req;
	struct crypto_ahash *tfm;
	int ret;
	static u32 flag = 0;

	tfm = crypto_alloc_ahash("hmac(sm3)", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", "sm3", PTR_ERR(tfm));
		return;
	}

	if (flag++ == 0)
		pr_info( "\ntesting of async %s (%s)\n", "sm3", get_driver_name(crypto_ahash, tfm));

	sg_init_table(&sg, 1);
	sg_set_buf(&sg, data, len);

	req = ahash_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("ahash request allocation failure\n");
		goto fail;
	}

	crypto_init_wait(&wait);
	ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);

	crypto_ahash_setkey(tfm, key, klen);
	ahash_request_set_crypt(req, &sg, out, len);

	ret = crypto_wait_req(crypto_ahash_digest(req), &wait);
	if (ret) {
		pr_err("hashing failed ret=%d\n", ret);
	}

	ahash_request_free(req);
fail:
	crypto_free_ahash(tfm);
}

static void hmac_sm3_step_test(const u8 *key, unsigned int klen, const u8 *data, unsigned int len, u8 *out)
{
	struct scatterlist sg;
	struct crypto_wait wait;
	struct ahash_request *req;
	struct crypto_ahash *tfm;
	int ret;
	unsigned int offset, step_len;
	static u32 flag = 0;

	tfm = crypto_alloc_ahash("hmac(sm3)", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", "sm3", PTR_ERR(tfm));
		return;
	}

	if (flag++ == 0)
		pr_info( "\ntesting of async %s (%s)\n", "sm3", get_driver_name(crypto_ahash, tfm));

	req = ahash_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("ahash request allocation failure\n");
		goto fail;
	}

	crypto_ahash_setkey(tfm, key, klen);
	crypto_init_wait(&wait);
	ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);

	ret = crypto_wait_req(crypto_ahash_init(req), &wait);
	if (ret) {
		pr_err("hashing failed ret=%d\n", ret);
		goto fail1;
	}

	for (offset = 0; offset < len;) {
		step_len = 0xFF;
		if (offset + step_len > len)
			step_len = len - offset;

		crypto_init_wait(&wait);
		ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);

		sg_init_table(&sg, 1);
		sg_set_buf(&sg, data + offset, step_len);
		ahash_request_set_crypt(req, &sg, out, step_len);

		ret = crypto_wait_req(crypto_ahash_update(req), &wait);
		if (ret) {
			pr_err("crypto_ahash_update failed ret=%d\n", ret);
			break;
		}
		offset += step_len;
	}

	crypto_init_wait(&wait);
	ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);

	sg_init_table(&sg, 1);
	sg_set_buf(&sg, data, 0);
	ahash_request_set_crypt(req, &sg, out, 0);
	ret = crypto_wait_req(crypto_ahash_final(req), &wait);
	if (ret) {
		pr_err("crypto_ahash_final failed ret=%d\n", ret);
	}

fail1:
	ahash_request_free(req);
fail:
	crypto_free_ahash(tfm);
}

static void sm4_ecb_enc_test(const u8 *key, const u8 *data, unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret;

	tfm = crypto_alloc_skcipher("ecb(sm4)", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n",  "ecb(sm4)", PTR_ERR(tfm));
		return;
	}

	if (flag++ == 0)
		pr_info( "\ntesting of async %s (%s) enc!\n", "ecb(sm4)", get_driver_name(crypto_skcipher, tfm));


	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", "ecb(sm4)");
		goto fail;
	}

	crypto_init_wait(&wait);
	skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	sg_init_table(&ssg, 1);
	sg_set_buf(&ssg, data, len);
	sg_init_table(&dsg, 1);
	sg_set_buf(&dsg, out, len);
	skcipher_request_set_crypt(req, &ssg, &dsg, len, NULL);

	ret = crypto_wait_req(crypto_skcipher_encrypt(req), &wait);
	if (ret) {
		pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);
}

static void sm4_ecb_enc_step_test(const u8 *key, const u8 *data, unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret;
	u32 offset, step_len;

	tfm = crypto_alloc_skcipher("ecb(sm4)", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", "ecb(sm4)", PTR_ERR(tfm));
		return;
	}

	if (flag++ == 0)
		pr_info( "\ntesting of async %s (%s) enc!\n", "ecb(sm4)", get_driver_name(crypto_skcipher, tfm));


	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", "ecb(sm4)");
		goto fail;
	}

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	for (offset = 0; offset < len;) {
		step_len = 256;
		if (offset + step_len > len)
			step_len = len - offset;

		crypto_init_wait(&wait);
		skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);

		sg_init_table(&ssg, 1);
		sg_set_buf(&ssg, data + offset, step_len);
		sg_init_table(&dsg, 1);
		sg_set_buf(&dsg, out + offset, step_len);
		skcipher_request_set_crypt(req, &ssg, &dsg, step_len, NULL);

		ret = crypto_wait_req(crypto_skcipher_encrypt(req), &wait);
		if (ret) {
			pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
			break;
		}
		offset += step_len;
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);
}

static void sm4_ecb_dec_test(const u8 *key, const u8 *data, unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret;

	tfm = crypto_alloc_skcipher("ecb(sm4)", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", "ecb(sm4)", PTR_ERR(tfm));
		return;
	}

	if (flag++ == 0)
		pr_info( "\ntesting of async %s (%s) dec!\n", "ecb(sm4)", get_driver_name(crypto_skcipher, tfm));


	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", "ecb(sm4)");
		goto fail;
	}

	crypto_init_wait(&wait);
	skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	sg_init_table(&ssg, 1);
	sg_set_buf(&ssg, data, len);
	sg_init_table(&dsg, 1);
	sg_set_buf(&dsg, out, len);
	skcipher_request_set_crypt(req, &ssg, &dsg, len, NULL);

	ret = crypto_wait_req(crypto_skcipher_decrypt(req), &wait);
	if (ret) {
		pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);
}

static void sm4_ecb_dec_step_test(const u8 *key, const u8 *data, unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret;
	u32 offset, step_len;

	tfm = crypto_alloc_skcipher("ecb(sm4)", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", "ecb(sm4)", PTR_ERR(tfm));
		return;
	}

	if (flag++ == 0)
		pr_info( "\ntesting of async %s (%s) dec!\n", "ecb(sm4)", get_driver_name(crypto_skcipher, tfm));


	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", "ecb(sm4)");
		goto fail;
	}

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	for (offset = 0; offset < len;) {
		step_len = 256;
		if (offset + step_len > len)
			step_len = len - offset;

		crypto_init_wait(&wait);
		skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);
		sg_init_table(&ssg, 1);
		sg_set_buf(&ssg, data + offset, step_len);
		sg_init_table(&dsg, 1);
		sg_set_buf(&dsg, out + offset, step_len);
		skcipher_request_set_crypt(req, &ssg, &dsg, step_len, NULL);

		ret = crypto_wait_req(crypto_skcipher_decrypt(req), &wait);
		if (ret) {
			pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
			break;
		}
		offset += step_len;
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);
}

static void sm4_cbc_enc_test(const u8 *key, u8 *iv, const u8 *data, unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret;

	tfm = crypto_alloc_skcipher("cbc(sm4)", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", "cbc(sm4)", PTR_ERR(tfm));
		return;
	}

	if (flag++ == 0)
		pr_info( "\ntesting of async %s (%s) enc!\n", "cbc(sm4)", get_driver_name(crypto_skcipher, tfm));


	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", "cbc(sm4)");
		goto fail;
	}

	crypto_init_wait(&wait);
	skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	sg_init_table(&ssg, 1);
	sg_set_buf(&ssg, data, len);
	sg_init_table(&dsg, 1);
	sg_set_buf(&dsg, out, len);
	skcipher_request_set_crypt(req, &ssg, &dsg, len, iv);

	ret = crypto_wait_req(crypto_skcipher_encrypt(req), &wait);
	if (ret) {
		pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);
}

static void sm4_cbc_enc_step_test(const u8 *key, u8 *iv, const u8 *data, unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret;
	u32 offset, step_len;

	tfm = crypto_alloc_skcipher("cbc(sm4)", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", "cbc(sm4)", PTR_ERR(tfm));
		return;
	}

	if (flag++ == 0)
		pr_info( "\ntesting of async %s (%s) enc!\n", "cbc(sm4)", get_driver_name(crypto_skcipher, tfm));


	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", "cbc(sm4)");
		goto fail;
	}

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	for (offset = 0; offset < len;) {
		step_len = 256;
		if (offset + step_len > len)
			step_len = len - offset;

		crypto_init_wait(&wait);
		skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);

		sg_init_table(&ssg, 1);
		sg_set_buf(&ssg, data + offset, step_len);
		sg_init_table(&dsg, 1);
		sg_set_buf(&dsg, out + offset, step_len);
		skcipher_request_set_crypt(req, &ssg, &dsg, step_len, iv);

		ret = crypto_wait_req(crypto_skcipher_encrypt(req), &wait);
		if (ret) {
			pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
			break;
		}
		offset += step_len;
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);
}

static void sm4_cbc_dec_test(const u8 *key, u8 *iv, const u8 *data, unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret;

	tfm = crypto_alloc_skcipher("cbc(sm4)", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", "cbc(sm4)", PTR_ERR(tfm));
		return;
	}

	if (flag++ == 0)
		pr_info( "\ntesting of async %s (%s) dec!\n", "cbc(sm4)", get_driver_name(crypto_skcipher, tfm));


	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", "cbc(sm4)");
		goto fail;
	}

	crypto_init_wait(&wait);
	skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	sg_init_table(&ssg, 1);
	sg_set_buf(&ssg, data, len);
	sg_init_table(&dsg, 1);
	sg_set_buf(&dsg, out, len);
	skcipher_request_set_crypt(req, &ssg, &dsg, len, iv);

	ret = crypto_wait_req(crypto_skcipher_decrypt(req), &wait);
	if (ret) {
		pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);
}

static void sm4_cbc_dec_step_test(const u8 *key, u8 *iv, const u8 *data, unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret;
	u32 offset, step_len;

	tfm = crypto_alloc_skcipher("cbc(sm4)", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", "cbc(sm4)", PTR_ERR(tfm));
		return;
	}

	if (flag++ == 0)
		pr_info( "\ntesting of async %s (%s) dec!\n", "cbc(sm4)", get_driver_name(crypto_skcipher, tfm));


	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", "cbc(sm4)");
		goto fail;
	}

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	for (offset = 0; offset < len;) {
		step_len = 256;
		if (offset + step_len > len)
			step_len = len - offset;

		crypto_init_wait(&wait);
		skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);

		sg_init_table(&ssg, 1);
		sg_set_buf(&ssg, data + offset, step_len);
		sg_init_table(&dsg, 1);
		sg_set_buf(&dsg, out + offset, step_len);
		skcipher_request_set_crypt(req, &ssg, &dsg, step_len, iv);

		ret = crypto_wait_req(crypto_skcipher_decrypt(req), &wait);
		if (ret) {
			pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
			break;
		}
		offset += step_len;
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);
}

static void sm4_ctr_enc_test(const u8 *key, u8 *iv, const u8 *data, unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret;

	tfm = crypto_alloc_skcipher("ctr(sm4)", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", "ctr(sm4)", PTR_ERR(tfm));
		return;
	}

	if (flag++ == 0)
		pr_info( "\ntesting of async %s (%s) enc!\n", "ctr(sm4)", get_driver_name(crypto_skcipher, tfm));


	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", "ctr(sm4)");
		goto fail;
	}

	crypto_init_wait(&wait);
	skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	sg_init_table(&ssg, 1);
	sg_set_buf(&ssg, data, len);
	sg_init_table(&dsg, 1);
	sg_set_buf(&dsg, out, len);
	skcipher_request_set_crypt(req, &ssg, &dsg, len, iv);

	ret = crypto_wait_req(crypto_skcipher_encrypt(req), &wait);
	if (ret) {
		pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);
}

static void sm4_ctr_enc_step_test(const u8 *key, u8 *iv, const u8 *data, unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret;
	u32 offset, step_len;

	tfm = crypto_alloc_skcipher("ctr(sm4)", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", "ctr(sm4)", PTR_ERR(tfm));
		return;
	}

	if (flag++ == 0)
		pr_info( "\ntesting of async %s (%s) enc!\n", "ctr(sm4)", get_driver_name(crypto_skcipher, tfm));


	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", "ctr(sm4)");
		goto fail;
	}

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	for (offset = 0; offset < len;) {
		step_len = 256;
		if (offset + step_len > len)
			step_len = len - offset;

		crypto_init_wait(&wait);
		skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);

		sg_init_table(&ssg, 1);
		sg_set_buf(&ssg, data + offset, step_len);
		sg_init_table(&dsg, 1);
		sg_set_buf(&dsg, out + offset, step_len);
		skcipher_request_set_crypt(req, &ssg, &dsg, step_len, iv);

		ret = crypto_wait_req(crypto_skcipher_encrypt(req), &wait);
		if (ret) {
			pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
			break;
		}
		offset += step_len;
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);
}

static void sm4_ctr_dec_test(const u8 *key, u8 *iv, const u8 *data, unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret;

	tfm = crypto_alloc_skcipher("ctr(sm4)", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", "ctr(sm4)", PTR_ERR(tfm));
		return;
	}

	if (flag++ == 0)
		pr_info( "\ntesting of async %s (%s) dec!\n", "ctr(sm4)", get_driver_name(crypto_skcipher, tfm));


	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", "ctr(sm4)");
		goto fail;
	}

	crypto_init_wait(&wait);
	skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	sg_init_table(&ssg, 1);
	sg_set_buf(&ssg, data, len);
	sg_init_table(&dsg, 1);
	sg_set_buf(&dsg, out, len);
	skcipher_request_set_crypt(req, &ssg, &dsg, len, iv);

	ret = crypto_wait_req(crypto_skcipher_decrypt(req), &wait);
	if (ret) {
		pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);
}

static void sm4_ctr_dec_step_test(const u8 *key, u8 *iv, const u8 *data, unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret;
	u32 offset, step_len;

	tfm = crypto_alloc_skcipher("ctr(sm4)", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", "ctr(sm4)", PTR_ERR(tfm));
		return;
	}

	if (flag++ == 0)
		pr_info( "\ntesting of async %s (%s) dec!\n", "ctr(sm4)", get_driver_name(crypto_skcipher, tfm));


	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", "ctr(sm4)");
		goto fail;
	}

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	for (offset = 0; offset < len;) {
		step_len = 256;
		if (offset + step_len > len)
			step_len = len - offset;

		crypto_init_wait(&wait);
		skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done, &wait);

		sg_init_table(&ssg, 1);
		sg_set_buf(&ssg, data + offset, step_len);
		sg_init_table(&dsg, 1);
		sg_set_buf(&dsg, out + offset, step_len);
		skcipher_request_set_crypt(req, &ssg, &dsg, step_len, iv);

		ret = crypto_wait_req(crypto_skcipher_decrypt(req), &wait);
		if (ret) {
			pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
			break;
		}
		offset += step_len;
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);
}

static void print_buf(u8 *name, u8 *buf, u32 len)
{
	u32 i;

	pr_info("%s[0x%x]:\n", name, len);
	for (i = 0; i < len; i += 16) {
		pr_info("%08x: %02x %02x %02x %02x %02x %02x %02x %02x "
		        "%02x %02x %02x %02x %02x %02x %02x %02x\n", i,
		        buf[i + 0], buf[i + 1], buf[i + 2], buf[i + 3], buf[i + 4], buf[i + 5], buf[i + 6], buf[i + 7],
		        buf[i + 8], buf[i + 9], buf[i + 10], buf[i + 11], buf[i + 12], buf[i + 13], buf[i + 14], buf[i + 15]);
	}
}

static void sm3_test(void)
{
	u8 *src;
	u8 kdgst[32], sdgst[32];
	u32 len;
	struct sm3_context ctx;
	struct sm3_func func_sm3;
	u64 time = jiffies;

	src = kmalloc(elen + 1024, GFP_KERNEL);
	if (src == NULL) {
		pr_err("%s:kmalloc failed!\n", __func__);
		return;
	}

	get_random_bytes(src, elen);
	sm3_func_alloc(&func_sm3);

	for (len = slen; len <= elen; len++) {
		sm3_dgst_test(src, len, kdgst);
		func_sm3.sm3_digest(&ctx, src, len, sdgst);
		if (memcmp(kdgst, sdgst, 32)) {
			pr_err("sm3_dgst_test err!\n");
			print_buf("src", src, len);
			print_buf("kdgst", kdgst, 32);
			print_buf("sdgst", sdgst, 32);
			break;
		}

		memset(kdgst, 0, 32);
		sm3_step_test(src, len, kdgst);
		if (memcmp(kdgst, sdgst, 32)) {
			pr_err("sm3_step_test err!\n");
			print_buf("src", src, len);
			print_buf("kdgst", kdgst, 32);
			print_buf("sdgst", sdgst, 32);
			break;
		}

		if (time + msecs_to_jiffies(100) < jiffies) {
			time = jiffies;
			schedule();
		}
	}

	kfree(src);
}

static void hmac_test(void)
{
	u8 *src;
	u8 kdgst[32], sdgst[32], key[64];
	u32 len;
	struct sm3_context ctx;
	struct sm3_func func_sm3;
	u64 time = jiffies;

	src = kmalloc(elen + 1024, GFP_KERNEL);
	if (src == NULL) {
		pr_err("%s:kmalloc failed!\n", __func__);
		return;
	}

	get_random_bytes(src, elen);
	get_random_bytes(key, 64);
	sm3_func_alloc(&func_sm3);

	func_sm3.hmac_sm3_setkey(&ctx, key, 64);

	for (len = slen; len <= elen; len++) {
		hmac_sm3_test(key, 64, src, len, kdgst);
		func_sm3.hmac_sm3_init(&ctx);
		func_sm3.hmac_sm3_finup(&ctx, src, len, sdgst);
		if (memcmp(kdgst, sdgst, 32)) {
			pr_err("hmac_sm3_test err!\n");
			print_buf("key", key, 64);
			print_buf("src", src, len);
			print_buf("kdgst", kdgst, 32);
			print_buf("sdgst", sdgst, 32);
			break;
		}
		hmac_sm3_step_test(key, 64, src, len, kdgst);
		if (memcmp(kdgst, sdgst, 32)) {
			pr_err("hmac_sm3_step_test err!\n");
			print_buf("key", key, 64);
			print_buf("src", src, len);
			print_buf("kdgst", kdgst, 32);
			print_buf("sdgst", sdgst, 32);
			break;
		}

		if (time + msecs_to_jiffies(100) < jiffies) {
			time = jiffies;
			schedule();
		}
	}

	kfree(src);
}

static void sm4_ecb_test(void)
{
	u8 *src, *kdst, *sdst;
	u8 key[16];
	u32 len;
	struct sm4_context ctx;
	struct sm4_func func_sm4;
	u64 time = jiffies;

	src = kmalloc(elen * 3, GFP_KERNEL);
	if (src == NULL) {
		pr_err("%s:kmalloc failed!\n", __func__);
		return;
	}
	kdst = src + elen;
	sdst = kdst + elen;

	get_random_bytes(src, elen);
	get_random_bytes(key, 16);
	sm4_func_alloc(&func_sm4);

	for (len = slen; len <= elen; len += 16) {
		sm4_ecb_enc_test(key, src, len, kdst);
		func_sm4.sm4_setkey_enc(&ctx, key);
		func_sm4.sm4_crypt_ecb_enc(&ctx, len, NULL, src, sdst);
		if (memcmp(kdst, sdst, len)) {
			pr_err("sm4_ecb_enc_test err!\n");
			print_buf("key", key, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			break;
		}
		sm4_ecb_enc_step_test(key, src, len, kdst);
		if (memcmp(kdst, sdst, len)) {
			pr_err("sm4_ecb_enc_step_test err!\n");
			print_buf("key", key, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			break;
		}

		sm4_ecb_dec_test(key, sdst, len, kdst);
		if (memcmp(kdst, src, len)) {
			pr_err("sm4_ecb_dec_test err!\n");
			print_buf("key", key, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			break;
		}
		sm4_ecb_dec_step_test(key, sdst, len, kdst);
		if (memcmp(kdst, src, len)) {
			pr_err("sm4_ecb_dec_step_test err!\n");
			print_buf("key", key, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			break;
		}

		if (time + msecs_to_jiffies(100) < jiffies) {
			time = jiffies;
			schedule();
		}
	}

	kfree(src);
}

static void sm4_cbc_test(void)
{
	u8 *src, *kdst, *sdst;
	u8 key[16], iv[16], initiv[16];
	u32 len;
	struct sm4_context ctx;
	struct sm4_func func_sm4;
	u64 time = jiffies;

	src = kmalloc(elen * 3, GFP_KERNEL);
	if (src == NULL) {
		pr_err("%s:kmalloc failed!\n", __func__);
		return;
	}
	kdst = src + elen;
	sdst = kdst + elen;

	get_random_bytes(src, elen);
	get_random_bytes(key, 16);
	get_random_bytes(initiv, 16);
	sm4_func_alloc(&func_sm4);

	for (len = slen; len <= elen; len += 16) {
		memcpy(iv, initiv, 16);
		sm4_cbc_enc_test(key, iv, src, len, kdst);

		memcpy(iv, initiv, 16);
		func_sm4.sm4_setkey_enc(&ctx, key);
		func_sm4.sm4_crypt_cbc_enc(&ctx, len, iv, src, sdst);
		if (memcmp(kdst, sdst, len)) {
			pr_err("sm4_cbc_enc_test err!\n");
			print_buf("key", key, 16);
			print_buf("initiv", initiv, 16);
			print_buf("iv", iv, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			break;
		}

		memcpy(iv, initiv, 16);
		sm4_cbc_enc_step_test(key, iv, src, len, kdst);
		if (memcmp(kdst, sdst, len)) {
			pr_err("sm4_cbc_enc_step_test err!\n");
			print_buf("key", key, 16);
			print_buf("initiv", initiv, 16);
			print_buf("iv", iv, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			break;
		}

		memcpy(iv, initiv, 16);
		sm4_cbc_dec_test(key, iv, sdst, len, kdst);
		if (memcmp(kdst, src, len)) {
			pr_err("sm4_cbc_dec_test err!\n");
			print_buf("key", key, 16);
			print_buf("initiv", initiv, 16);
			print_buf("iv", iv, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			break;
		}

		memcpy(iv, initiv, 16);
		sm4_cbc_dec_step_test(key, iv, sdst, len, kdst);
		if (memcmp(kdst, src, len)) {
			pr_err("sm4_cbc_dec_step_test err!\n");
			print_buf("key", key, 16);
			print_buf("initiv", initiv, 16);
			print_buf("iv", iv, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			break;
		}

		if (time + msecs_to_jiffies(100) < jiffies) {
			time = jiffies;
			schedule();
		}
	}

	kfree(src);
}

static void sm4_ctr_test(void)
{
	u8 *src, *kdst, *sdst;
	u8 key[16], iv[16], initiv[16];
	u32 len;
	struct sm4_context ctx;
	struct sm4_func func_sm4;
	u64 time = jiffies;

	src = kmalloc(elen * 3, GFP_KERNEL);
	if (src == NULL) {
		pr_err("%s:kmalloc failed!\n", __func__);
		return;
	}
	kdst = src + elen;
	sdst = kdst + elen;

	get_random_bytes(src, elen);
	get_random_bytes(key, 16);
	get_random_bytes(initiv, 16);
	sm4_func_alloc(&func_sm4);

	for (len = slen; len <= elen; len += 16) {
		memcpy(iv, initiv, 16);
		sm4_ctr_enc_test(key, iv, src, len, kdst);

		memcpy(iv, initiv, 16);
		func_sm4.sm4_setkey_enc(&ctx, key);
		func_sm4.sm4_crypt_ctr(&ctx, len, iv, src, sdst);
		if (memcmp(kdst, sdst, len)) {
			pr_err("sm4_ctr_enc_test err!\n");
			print_buf("key", key, 16);
			print_buf("initiv", initiv, 16);
			print_buf("iv", iv, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			break;
		}

		memcpy(iv, initiv, 16);
		sm4_ctr_enc_step_test(key, iv, src, len, kdst);
		if (memcmp(kdst, sdst, len)) {
			pr_err("sm4_ctr_enc_step_test err!\n");
			print_buf("key", key, 16);
			print_buf("initiv", initiv, 16);
			print_buf("iv", iv, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			break;
		}

		memcpy(iv, initiv, 16);
		sm4_ctr_dec_test(key, iv, sdst, len, kdst);
		if (memcmp(kdst, src, len)) {
			pr_err("sm4_ctr_dec_test err!\n");
			print_buf("key", key, 16);
			print_buf("initiv", initiv, 16);
			print_buf("iv", iv, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			break;
		}

		memcpy(iv, initiv, 16);
		sm4_ctr_dec_step_test(key, iv, sdst, len, kdst);
		if (memcmp(kdst, src, len)) {
			pr_err("sm4_ctr_dec_step_test err!\n");
			print_buf("key", key, 16);
			print_buf("initiv", initiv, 16);
			print_buf("iv", iv, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			break;
		}

		if (time + msecs_to_jiffies(100) < jiffies) {
			time = jiffies;
			schedule();
		}
	}

	kfree(src);
}

void print_help(void)
{
	pr_info("------BAD PARAM!------\n");
	pr_info("alg: \"sm3\", \"hmac(sm3)\", \"ecb(sm4)\", \"cbc(sm4)\", \"ctr(sm4)\"\n");
	pr_info("slen: test min length\n");
	pr_info("elen: test max length\n");
	pr_info("------BAD PARAM!------\n");
}

static int sm34_test_init(void)
{
	u32 len;

	if (slen > elen) {
		len = slen;
		slen = elen;
		elen = len;
	}

	if (alg == NULL) {
		print_help();
		return -1;
	}

	if (memcmp(alg, "sm3", strlen("sm3")) == 0) {
		sm3_test();
		return -1;
	}
	if (memcmp(alg, "hmac(sm3)", strlen("hmac(sm3)")) == 0) {
		hmac_test();
		return -1;
	}
	if (memcmp(alg, "ecb(sm4)", strlen("ecb(sm4)")) == 0) {
		slen = (slen + 0xF) & 0xFFFFFFF0;
		elen = (elen + 0xF) & 0xFFFFFFF0;
		sm4_ecb_test();
		return -1;
	}
	if (memcmp(alg, "cbc(sm4)", strlen("cbc(sm4)")) == 0) {
		slen = (slen + 0xF) & 0xFFFFFFF0;
		elen = (elen + 0xF) & 0xFFFFFFF0;
		sm4_cbc_test();
		return -1;
	}
	if (memcmp(alg, "ctr(sm4)", strlen("ctr(sm4)")) == 0) {
		slen = (slen + 0xF) & 0xFFFFFFF0;
		elen = (elen + 0xF) & 0xFFFFFFF0;
		sm4_ctr_test();
		return -1;
	}

	print_help();

	return -1;
}

static void sm34_test_exit(void)
{
}

module_init(sm34_test_init);
module_exit(sm34_test_exit);
MODULE_DESCRIPTION("PHYTIUM TEST");
MODULE_LICENSE("GPL");
